/* NiuTrans.Tensor - an open-source tensor library
* Copyright (C) 2017, Natural Language Processing Lab, Northestern University.
* All rights reserved.
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*   http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/

/*
* $Created by: LI Yinqiao (li.yin.qiao.2012@hotmail.com) 2018-06-14
*/

#include "../../XTensor.h"
#include "../../XDevice.h"
#include "ConvertDataType.cuh"

namespace nts { // namespace nts(NiuTrans.Tensor)

#ifdef USE_CUDA
#include <cuda_runtime.h>
#include <cublas_v2.h>
#include <cuda_fp16.h>
#include <stdint.h>

__global__ 
void KernelFloatToFloat16(float * s, __half * t, int size)
{
    int i = blockDim.x * blockIdx.x + threadIdx.x;

    if (i < size){
        t[i] = __float2half(s[i]);
    }
}

__global__ 
void KernelFloat16ToFloat(__half * s, float * t, int size)
{
    int i = blockDim.x * blockIdx.x + threadIdx.x;

    if (i < size){
        t[i] = __half2float(s[i]);
    }
}

__global__ 
void KernelFloatToInt(float * inputData, int * outputData, int size)
{
    int i = blockDim.x * blockIdx.x + threadIdx.x;

    if (i < size){
        outputData[i] = (int)(inputData[i]);
    }
}

__global__
void KernelFloatToInt8(float * inputData, int8_t * outputData, int size)
{
    int i = blockDim.x * blockIdx.x + threadIdx.x;

    if (i < size) {
        outputData[i] = (int8_t)(inputData[i]);
    }
}

__global__ 
void KernelIntToFloat(int * inputData, float * outputData, int size)
{
    int i = blockDim.x * blockIdx.x + threadIdx.x;

    if (i < size){
        outputData[i] = (float)(inputData[i]);
    }
}

__global__
void KernelIntToFloat8(int8_t * inputData, float * outputData, int size)
{
    int i = blockDim.x * blockIdx.x + threadIdx.x;

    if (i < size) {
        outputData[i] = (float)(inputData[i]);
    }
}

/* 
data conversion (cuda code) 
>> devID - device id
>> s - source data array
>> typeS - source data type
>> t - target data array
>> typeT - target data type
>> size - number of the items in s (and t)
*/
void _CudaConvertDataType(int devID, void * s, TENSOR_DATA_TYPE typeS, void * t, TENSOR_DATA_TYPE typeT, int size)
{
    CheckNTErrors((devID >= 0), "This code must be run on GPUs!");

    if(typeS == typeT)
        return;

    int gridSize[3];
    int blockSize[3];

    GDevs.GetCudaThread(devID, size, gridSize, blockSize);

    dim3 blocks(gridSize[0]);
    dim3 threads(blockSize[0]);

    int devIDBackup;
    ProtectCudaDev(devID, devIDBackup);

    if(typeS == X_FLOAT && typeT == X_FLOAT16)
        KernelFloatToFloat16<<<blocks, threads>>>((float*)s, (__half*)t, size);
    else if(typeS == X_FLOAT16 && typeT == X_FLOAT)
        KernelFloat16ToFloat<<<blocks, threads>>>((__half*)s, (float*)t, size);
    else{
        ShowNTErrors("Unsupported data types for conversion!");
    }

    ProtectCudaDev(devID, devIDBackup);
}

/*
convert data type (cuda code) 
>> input - input tensor
>> output - output tensor
*/
void _CudaConvertDataType(const XTensor * input, XTensor * output)
{
    //CheckNTErrors((input->unitSize == output->unitSize), "Input and Output must be same in size!");

    if (input->dataType == output->dataType)
        return;

    int gridSize[3];
    int blockSize[3];

    GDevs.GetCudaThread(input->devID, input->unitNum, gridSize, blockSize);
    dim3 blocks(gridSize[0]);
    dim3 threads(blockSize[0]);

    int devIDBackup;
    ProtectCudaDev(input->devID, devIDBackup);

    if(input->dataType == X_FLOAT && output->dataType == X_INT)
        KernelFloatToInt<<<blocks, threads>>>((float*)input->data, (int*)output->data, input->unitNum);
    else if(input->dataType == X_INT && output->dataType == X_FLOAT)
        KernelIntToFloat<<<blocks, threads>>>((int*)input->data, (float*)output->data, input->unitNum);
    else if (input->dataType == X_FLOAT && output->dataType == X_FLOAT16)
        KernelFloatToFloat16 << <blocks, threads >> >((float*)input->data, (__half*)output->data, input->unitNum);   
    else if (input->dataType == X_FLOAT16 && output->dataType == X_FLOAT)
        KernelFloat16ToFloat << <blocks, threads >> >((__half*)input->data, (float*)output->data, input->unitNum); 
    else if (input->dataType == X_FLOAT && output->dataType == X_INT8)
        KernelFloatToInt8 << <blocks, threads >> >((float*)input->data, (int8_t*)output->data, input->unitNum);
    else if (input->dataType == X_INT8 && output->dataType == X_FLOAT)
        KernelIntToFloat8 << <blocks, threads >> >((int8_t*)input->data, (float*)output->data, input->unitNum);
    else{
        ShowNTErrors("Unsupported data types for conversion!");
    }

    ProtectCudaDev(input->devID, devIDBackup);
}

__global__
void KernelConvertf32tof16(float* input, __half *output, int stride, int strideNum, int blockNum, int size)
{
    int idx = blockDim.x * blockIdx.x + threadIdx.x;
    int idy = blockDim.y * blockIdx.y + threadIdx.y;
    int blockIndex = idy / stride;
    int offsetInBlock = idy% stride;
#pragma unroll
    for (int i = idx * stride + stride * strideNum * blockIndex + offsetInBlock;
        i < stride * strideNum * blockIndex + offsetInBlock + stride * strideNum && i < size;
        i += stride * blockDim.x) {
        output[i] = __float2half(input[i]);
    }
}
__global__
void KernelConvertf16tof32(__half* input, float *output, int stride, int strideNum, int blockNum, int size)
{
    int idx = blockDim.x * blockIdx.x + threadIdx.x;
    int idy = blockDim.y * blockIdx.y + threadIdx.y;
    int blockIndex = idy / stride;
    int offsetInBlock = idy% stride;
#pragma unroll
    for (int i = idx * stride + stride * strideNum * blockIndex + offsetInBlock;
        i < stride * strideNum * blockIndex + offsetInBlock + stride * strideNum && i < size;
        i += stride * blockDim.x) {
        output[i] = __half2float(input[i]);
    }
}

/*
convert data type (cuda code)
>> input - input tensor
>> output - output tensor
*/
void _CudaConvertDataTypeNew(const XTensor * input, XTensor * output)
{
    //CheckNTErrors((input->unitSize == output->unitSize), "Input and Output must be same in size!");
    if (input->dataType == output->dataType)
        return;
    //int dimRDI = input->order - 1;
    int dimRDI = 0;
    int stride = 1;
    int strideNumA = input->dimSizeRDI[dimRDI];
    for (int i = 0; i < dimRDI; i++)
        stride *= input->dimSizeRDI[i];

    int blockNum = 1;
    for (int i = dimRDI + 1; i < input->order; i++)
        blockNum *= input->dimSizeRDI[i];

    int gridSize[3];
    int blockSize[3];
    /*for (int i = 0; i < input->order; ++i)
    {
        printf("%d ", input->dimSizeRDI[i]);
    }
    printf("\n");
    printf("%d %d\n", dimRDI, input->dimSizeRDI[dimRDI]);
    printf("%d %d %d\n", stride, strideNumA, blockNum);*/
    int workerNum = 64;
    GDevs.GetCudaThread2D(input->devID, workerNum, stride * blockNum, MAX_INT, gridSize, blockSize);
    dim3 blocks(gridSize[0], gridSize[1]);
    dim3 threads(blockSize[0], blockSize[1]);

    int devIDBackup;
    ProtectCudaDev(input->devID, devIDBackup);

    if (input->dataType == X_FLOAT && output->dataType == X_INT)
        KernelFloatToInt << <blocks, threads >> > ((float*)input->data, (int*)output->data, input->unitNum);
    else if (input->dataType == X_INT && output->dataType == X_FLOAT)
        KernelIntToFloat << <blocks, threads >> > ((int*)input->data, (float*)output->data, input->unitNum);
    else if (input->dataType == X_FLOAT && output->dataType == X_FLOAT16)
        KernelConvertf32tof16 << <blocks, threads >> > ((float*)input->data, (__half*)output->data, stride, strideNumA, blockNum, input->unitNum);
    else if (input->dataType == X_FLOAT16 && output->dataType == X_FLOAT)
        KernelConvertf16tof32 << <blocks, threads >> >((__half*)input->data, (float*)output->data, stride, strideNumA, blockNum, input->unitNum);
    else {
        ShowNTErrors("Unsupported data types for conversion!");
    }

    ProtectCudaDev(input->devID, devIDBackup);
}




#endif // USE_CUDA

} // namespace nts(NiuTrans.Tensor)